/*
 * Timer/counter unit support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "tcu-x1600.h"



// Register locations.

enum Regs : unsigned
{
  Tcu_control_base                  = 0x04c,      // TCRn

  // Block size/step/offset for the above register set.

  Tcu_data_block_offset             = 0x010,
};

enum Regs_x1600 : unsigned
{
  // X1600 channel-related locations.

  Tcu_store_flag_status             = 0x200,      // TSFR
  Tcu_store_set_flag                = 0x204,      // TSFSR
  Tcu_store_clear_flag              = 0x208,      // TSFCR
  Tcu_store_mask_status             = 0x210,      // TSMR
  Tcu_store_set_mask                = 0x214,      // TSMSR
  Tcu_store_clear_mask              = 0x218,      // TSMCR

  Tcu_capture_control_base          = 0x0c0,      // TCRn
  Tcu_capture_counter_base          = 0x0e0,      // CAPVRn
  Tcu_filter_value_base             = 0x1a0,      // FIRVRn
  Tcu_stored_counter_base           = 0x220,      // TSVRn
  Tcu_stored_filter_counter_base    = 0x240,      // TSFVRn

  // Block size/step/offset for the above register set.

  Tcu_capture_block_offset          = 0x004,

  // Limits.

  Tcu_capture_channel_max           = 3,
};

// Field definitions.

// Flag/mask register bits.

enum Flag_bit_numbers : unsigned
{
  Half_match_wdt                    = 24,         // HFLAGW

  // Flag/mask group bit offsets.

  Half_match_shift                  = 16,
  Full_match_shift                  = 0,
};

// Counter data constraints.

enum Data_masks : unsigned
{
  Data_mask                         = 0xffff,
};

enum Control_bits_x1600 : unsigned
{
  Store_negative_edge_enable        = 0x04000000, // STORE_NEG_EN
  Store_positive_edge_enable        = 0x02000000, // STORE_POS_EN
  Store_enable                      = 0x01000000, // STORE_EN

  Count_mode_field_mask             = 0x3,        // COUNT_MODE
  Count_mode_wrap_at_full_data      = 0,
  Count_mode_wrap_at_field_limit    = 1,
  Count_mode_stop_at_field_limit    = 2,
  Count_mode_field_shift            = 22,

  Count_source_field_mask           = 0x3f,
  Count_gpio1_negative_edge_enable  = 0x20,       // GPIO1_NEG_EN
  Count_gpio1_positive_edge_enable  = 0x10,       // GPIO1_POS_EN
  Count_gpio0_negative_edge_enable  = 0x08,       // GPIO0_NEG_EN
  Count_gpio0_positive_edge_enable  = 0x04,       // GPIO0_POS_EN
  Count_clock_negative_edge_enable  = 0x02,       // CLK_NEG_EN
  Count_clock_positive_edge_enable  = 0x01,       // CLK_POS_EN
  Count_source_field_shift          = 16,

  Count_shutdown                    = 0x00008000, // SHUTDOWN
  Count_gate_polarity_low           = 0x00000000, // GATE_POLA = 0
  Count_gate_polarity_high          = 0x00004000, // GATE_POLA = 1
  Count_direction_polarity_low      = 0x00000000, // DIRECTION_POLA = 0
  Count_direction_polarity_high     = 0x00002000, // DIRECTION_POLA = 1

  Count_gate_select_field_mask      = 0x3,        // GATE_SEL
  Count_gate_select_none            = 0,
  Count_gate_select_clock           = 1,
  Count_gate_select_gpio0           = 2,
  Count_gate_select_gpio1           = 3,
  Count_gate_select_field_shift     = 11,

  Count_direction_select_field_mask  = 0x7,        // DIRECTION_SEL
  Count_direction_select_none        = 0,
  Count_direction_select_clock       = 1,
  Count_direction_select_gpio0       = 2,
  Count_direction_select_gpio1       = 3,
  Count_direction_select_gpio01      = 4,
  Count_direction_select_field_shift = 8,

  Count_gpio1_enable                = 0x00000080, // GPIO1_EN
  Count_gpio0_enable                = 0x00000040, // GPIO0_EN
};

enum Capture_bits : unsigned
{
  Capture_select_mask               = 0x00070000, // CAPTURE_SEL
  Capture_select_clock              = 0x00000000, // CAPTURE_SEL = 0
  Capture_select_gpio0              = 0x00010000, // CAPTURE_SEL = 1
  Capture_select_gpio1              = 0x00020000, // CAPTURE_SEL = 2
  Capture_count_mask                = 0x000000ff, // CAPTURE_NUM
};

enum Capture_value_bits : unsigned
{
  Capture_time_all_mask             = 0xffff0000, // CAPTURE_ALL
  Capture_time_high_mask            = 0x0000ffff, // CAPTURE_HEIGHT (sic)
};

enum Filter_value_bits : unsigned
{
  Filter_gpio1_value_mask           = 0x03ff0000, // FIL_B
  Filter_gpio0_value_mask           = 0x000003ff, // FIL_A
};

enum Store_filter_bits : unsigned
{
  Store_filter_value_mask           = 0x3ff,      // STR_FIL
};



// Channel abstraction.

Tcu_x1600_channel::Tcu_x1600_channel(l4_addr_t addr, uint8_t channel)
: Tcu_channel(addr, channel)
{
}

// Operation methods.

void
Tcu_x1600_channel::enable()
{
  // NOTE: Use a positive clock edge for the timer event by default.

  set_field(Tcu_control_base + _channel * Tcu_data_block_offset,
            Count_source_field_mask, Count_source_field_shift,
            Count_clock_positive_edge_enable);

  Tcu_channel::enable();
}

uint8_t
Tcu_x1600_channel::get_count_mode()
{
  return get_field(Tcu_control_base + _channel * Tcu_data_block_offset,
                   Count_mode_field_mask, Count_mode_field_shift);
}

void
Tcu_x1600_channel::set_count_mode(uint8_t mode)
{
  set_field(Tcu_control_base + _channel * Tcu_data_block_offset,
            Count_mode_field_mask, Count_mode_field_shift, mode);
}



// Peripheral abstraction.

Tcu_x1600_chip::Tcu_x1600_chip(l4_addr_t start, l4_addr_t end)
: Tcu_chip(start, end)
{
}

Tcu_channel *Tcu_x1600_chip::_get_channel(l4_addr_t addr, uint8_t channel)
{
  return new Tcu_x1600_channel(addr, channel);
}



// C language interface functions.

void *x1600_tcu_init(l4_addr_t tcu_base, l4_addr_t tcu_base_end)
{
  return (void *) new Tcu_x1600_chip(tcu_base, tcu_base_end);
}

void *x1600_tcu_get_channel(void *tcu, uint8_t channel)
{
  return static_cast<Tcu_x1600_chip *>(tcu)->get_channel(channel);
}

void x1600_tcu_disable(void *tcu_channel)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->disable();
}

void x1600_tcu_enable(void *tcu_channel)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->enable();
}

int x1600_tcu_is_enabled(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->is_enabled();
}

uint8_t x1600_tcu_get_clock(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_clock();
}

void x1600_tcu_set_clock(void *tcu_channel, uint8_t clock)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_clock(clock);
}

uint32_t x1600_tcu_get_prescale(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_prescale();
}

void x1600_tcu_set_prescale(void *tcu_channel, uint32_t prescale)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_prescale(prescale);
}

uint32_t x1600_tcu_get_counter(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_counter();
}

void x1600_tcu_set_counter(void *tcu_channel, uint32_t value)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_counter(value);
}

uint8_t x1600_tcu_get_count_mode(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_count_mode();
}

void x1600_tcu_set_count_mode(void *tcu_channel, uint8_t mode)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_count_mode(mode);
}

uint32_t x1600_tcu_get_full_data_value(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_full_data_value();
}

void x1600_tcu_set_full_data_value(void *tcu_channel, uint32_t value)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_full_data_value(value);
}

uint32_t x1600_tcu_get_half_data_value(void *tcu_channel)
{
  return static_cast<Tcu_x1600_channel *>(tcu_channel)->get_half_data_value();
}

void x1600_tcu_set_half_data_value(void *tcu_channel, uint32_t value)
{
  static_cast<Tcu_x1600_channel *>(tcu_channel)->set_full_data_value(value);
}
